default = '_def'
default_key = '_defkey'
tags_key = '_tags'

def ask(word: str, path=()):
    print(path[0], end='')
    for dir in path[1:]:
        print(f', {dir}', end='')
    print(f' для слова "{word}": ', end='')
    return input()


from .Util.Util import extract_all, rel_format
from . import Forms
import random
import sys
import yaml


class WordList():

    word_types = ('безл.', )
    word_traits = ('абр', 'пинг', 'перс')

    def __init__(self):
        self.forms = dict()
        self.tags = dict()
        self.__cache = []

    def delete(self, *args):
        for name in args:
            del self.forms[name]
            del self.tags[name]

    def get(self, *args):
        tags = list(args)

        forms = extract_all(tags, 'ед.ч.', 'мн.ч.', 'и.п.', 'р.п.', 'д.п.', 'в.п.', 'т.п.', 'п.п.', 'п.в.', 'н.в.', 'б.в.', '1л', '2л', '3л', to_pop=True)

        if 'гл.' in args:
            forms.extend(extract_all(tags, 'инф.', 'пов.', 'м.р.', 'с.р.', 'ж.р.', to_pop=True))
        elif 'пр.' in args:
            forms.extend(extract_all(tags, 'одуш.', 'неодуш.', 'м.р.', 'с.р.', 'ж.р.', to_pop=True))

        tag_correlations = tuple(filter(
            lambda word_tags: all(tag in word_tags[1] for tag in tags), self.tags.items()
            )) if len(tags) > 0 else tuple(self.tags.items())

        if len(tag_correlations) == 0:
            return

        word = random.choice(tag_correlations)[0]

        return rel_format(word, Forms.lookup(self.forms[word], *forms))

    def read(self, *, tagsin=sys.stdin, formsin=sys.stdin):
        if tagsin is not None and formsin is not None:
            saved_forms = yaml.load(formsin, yaml.Loader)
            if type(saved_forms) == dict:
                self.forms = saved_forms

            saved_tags = yaml.load(tagsin, yaml.Loader)
            if type(saved_tags) == dict:
                self.tags = saved_tags

    def flush(self, *, tagsout=sys.stdout, formsout=sys.stdout):
        if len(self.__cache) == 0:
            return

        cached_tags = {}
        cached_forms = {}
        for key in self.__cache:
            cached_tags[key] = self.tags[key]
            cached_forms[key] = self.forms[key]

        yaml.dump(cached_tags, tagsout, indent=2, allow_unicode=True)
        yaml.dump(cached_forms, formsout, indent=2, allow_unicode=True)

        self.__cache.clear()

    def insert(self, name, tags):
        types = extract_all(tags, *self.word_types, to_pop=True)
        for word_type in types:
            match word_type:
                case 'безл.':
                    tags.extend(('м.р.', 'ж.р.'))

        traits = extract_all(tags, *self.word_traits, to_pop=False)
        for word_trait in traits:
            match word_trait:
                case 'абр' | 'пинг':
                    tags.append('нескл.')
                case 'перс':
                    tags.append('ед.ч.')

        print(f'Базовая форма слова: "{name}".')
        print('Часть речи: ', end='')
        if 'гл.' in tags:
            print('глагол.')
        elif 'пр.' in tags:
            print('прилагательное/причастие.')
        elif 'сущ.' in tags:
            print('существительное.')
        else:
            print('другое (наречие/деепричастие/...)')
        print(f'Теги: {tags}.')

        if name in self.tags:
            print(f'!! Это слово уже есть в списке.')
            print(f'   Его теги: {self.tags[name]}.')
            return
        if 'гл.' in tags:
            if 'сов.' not in tags and 'несов.' not in tags:
                print('!! Пожалуйста, укажите завершённость глагола: совершённый "сов." или несовершённый "несов."')
                return
        if 'сущ.' in tags:
            if 'м.р.' not in tags and 'с.р.' not in tags and 'ж.р.' not in tags:
                print('! Вы не указали род существительного (м.р.|с.р.|ж.р.|безл.).')
                print('  Выборка в тексте в основном требует род, что значит, ваше слово практически не будет показываться.')
        if len(name.split()) > 1:
            print('! Крайне не рекомендуем добавлять словосочетания.')
        if len(tags) == 0:
            print('! Теги добавляются при вводе. Например, "любить гл. несов."; "человек сущ. м.р."')

        print('Продолжить? (пустой ввод, если да) ', end='')
        if len(input()) != 0:
            return
        print()

        self.forms[name] = Forms.construct_forms(name, *tags)
        self.tags[name] = tags
        self.__cache.append(name)



from .FormConstructors import Adj, Noun, Verb, Defines


def lookup(self, *args):
    if type(self) is not dict:
        return self

    default_key = Defines.default if Defines.default_key not in self else self[Defines.default_key]

    for key, value in self.items():
        if key in args:
            return lookup(value, *args)

    value = self[default_key]
    return lookup(value, *args)


def construct_forms(word: str, *args, path=()):
    if 'гл.' in args:
        return {
            Defines.default: Verb.conjugable(word, *args, path=path),
            'пов.': Verb.imperative(word, path=path + ('Повелительное наклонение', )),
            'инф.': '+'
        }
    elif 'сущ.' in args:
        return '+' if 'нескл.' in args else Noun.declined(word, *args, path=path)
    elif 'пр.' in args:
        return {
            Defines.default: Adj.singular(word, path=path + ('Единственное число', )),
            'мн.ч.': Adj.plural(word, path=path + ('Множественное число', ))
        }
    else:
        return '+'


from . import Defines


def singular(word: str, *, path=()):
    mascul_neu_preset = {
        'р.п.': Defines.ask(word, path=path + ('мужской/средний род', 'родительный падеж')),
        'д.п.': Defines.ask(word, path=path + ('мужской/средний род', 'дательный падеж')),
        'т.п.': Defines.ask(word, path=path + ('мужской/средний род', 'творительный падеж')),
        'п.п.': Defines.ask(word, path=path + ('мужской/средний род', 'предложный падеж'))
    }

    mascul_preset = mascul_neu_preset.copy()
    mascul_preset[Defines.default] = '+'
    mascul_preset['в.п.'] = {
        Defines.default: mascul_preset[Defines.default],
        'одуш.': mascul_preset['р.п.']
    }

    neu_preset = mascul_neu_preset.copy()
    neu_preset[Defines.default] = Defines.ask(word, path=path + ('средний род', 'именительный падеж'))
    neu_preset['в.п.'] = {
        Defines.default: neu_preset[Defines.default],
        'одуш.': neu_preset['р.п.']
    }

    fem_case = Defines.ask(word, path=path + ('женский род', 'родительный/дательный/творительный/предложный падеж'))

    return {
        Defines.default: mascul_preset,
        'с.р.': neu_preset,
        'ж.р.': {
            Defines.default: Defines.ask(word, path=path + ('женский род', 'именительный падеж')),
            'р.п.': fem_case,
            'д.п.': fem_case,
            'в.п.': Defines.ask(word, path=path + ('женский род', 'винительный падеж')),
            'т.п.': fem_case,
            'п.п.': fem_case
        }
    }

def plural(word: str, *, path=()):
    nominative = Defines.ask(word, path=path + ('именительный падеж', ))
    genitive = Defines.ask(word, path=path + ('родительный падеж', ))
    return {
        Defines.default: nominative,
        'р.п.': genitive,
        'д.п.': Defines.ask(word, path=path + ('дательный падеж', )),
        'в.п.': {
            Defines.default: nominative,
            'одуш.': genitive
        },
        'т.п.': Defines.ask(word, path=path + ('творительный падеж', )),
        'п.п.': Defines.ask(word, path=path + ('предложный падеж', )),
    }


from . import Defines


def imperative(word: str, *, path=()):
    imperative_form = Defines.ask(word, path)
    return {
        Defines.default: imperative_form,
        'мн.ч.': imperative_form + 'те'
    }

def conjugable(word: str, *args, path=()):
    forms = {}

    forms[Defines.default] = '+'
    forms['п.в.'] = time_past(word, path=path + ('Прошедшее время', ))
    forms['б.в.'] = time(word, *args, 'б.в.', path=path + ('Будущее время', ))
    if 'несов.' in args:
        forms['н.в.'] = time(word, *args, 'н.в.', path=path + ('Настоящее время', ))

    return forms

def time_past(word: str, *, path=()):
    forms = {}

    standard_form = Defines.ask(word, path)

    forms[Defines.default] = standard_form  # мужской род
    if standard_form[-1] != 'л':
        standard_form = standard_form + 'л'
    forms['ж.р.'] = standard_form + 'а'
    forms['с.р.'] = standard_form + 'о'
    forms['мн.ч.'] = standard_form + 'и'

    return forms

def time(word: str, *args, path=()):
    return {
        Defines.default: '+',
        '1л': person(word, *args, '1л', path=path + ('первое лицо', )),
        '2л': person(word, *args, '2л', path=path + ('второе лицо', )),
        '3л': person(word, *args, '3л', path=path + ('третье лицо', ))
    }

def person(word: str, *args, path=()):
    if 'б.в.' in args and 'несов.' in args:
        return {
            Defines.default: 'буду +',
            'мн.ч.': 'будем +'
        } if '1л' in args else {
            Defines.default: 'будешь +',
            'мн.ч.': 'будете +'
        } if '2л' in args else {
            Defines.default: 'будет +',
            'мн.ч.': 'будут +'
        }

    return {
        Defines.default: Defines.ask(word, path=path + ('единственное число', )),
        'мн.ч.': Defines.ask(word, path=path + ('множественное число', ))
    }


from . import Defines


def declined(word: str, *args, path=()):
    return {
        Defines.default: case(word, *args, 'и.п.', path=path + ('Именительный падеж', )),
        'р.п.': case(word, *args, 'р.п.', path=path + ('Родительный падеж', )),
        'д.п.': case(word, *args, 'д.п.', path=path + ('Дательный падеж', )),
        'в.п.': case(word, *args, 'в.п.', path=path + ('Винительный падеж', )),
        'т.п.': case(word, *args, 'т.п.', path=path + ('Творительный падеж', )),
        'п.п.': case(word, *args, 'п.п.', path=path + ('Предложный падеж', ))
    }

def case(word: str, *args, path=()):
    return {
        Defines.default: '+' if 'и.п.' in args else Defines.ask(word, path)
    } if 'ед.ч.' in args or 'мн.ч.' in args else {
        Defines.default: '+' if 'и.п.' in args else Defines.ask(word, path + ('единственное число', )),
        'мн.ч.': Defines.ask(word, path + ('множественное число', ))
    }


"""A Russian text generator based on templating texts and word picking"""

__version__ = "0.4.1"

from WordDict.WordList import WordList
import os.path


def user_interact(db):
    print('Введите слова:')
    while True:
        print()
        word = input()
        if len(word) == 0:
            break
        tokens = word.split(sep=' ')
        name = tokens[0]
        tags = tokens[1:] if len(tokens) > 1 else ()
        db.insert(name, tags)

    print('Введите теги:')
    while True:
        print()
        tags = input()
        if len(tags) == 0:
            break
        print(db.get(*tags.split()))


def main():
    db = WordList()

    forms_db = "forms.yaml"
    tags_db = "tags.yaml"

    if os.path.isfile(forms_db) and os.path.isfile(tags_db):
        with open(forms_db, mode='r', encoding='utf-8') as formsin:
            with open(tags_db, mode='r', encoding='utf-8') as tagsin:
                db.read(formsin=formsin, tagsin=tagsin)

    try:
        user_interact(db)
    finally:
        with open(forms_db, mode='a', encoding='utf-8') as formsout:
            with open(tags_db, mode='a', encoding='utf-8') as tagsout:
                db.flush(formsout=formsout, tagsout=tagsout)


if __name__ == '__main__':
    main()
